:- module(abstract, [abstract_program/2]).
:- use_module(library(assoc)).
:- use_module(library(lists)).
:- use_module(library(terms)).
:- use_module(library(ugraphs)).
:- use_module(polyhedra).
:- use_module(utilities).

:- dynamic head_to_normal_clause/2.
:- dynamic head_to_linear_clause/3.
:- dynamic head_to_ground_clause/2.
:- dynamic new_clause/1.

abstract_program(File, Linear_SCCs) :-
	concat(['../', File, '/', File, '.pl'], PathFile),
	statistics(runtime, _),
	format("~n{parsing, ", []), ttyflush,
	parse_program(PathFile, Clauses),
	format("normalising, ", []), ttyflush,
	normal_program(Clauses, Normal_Clauses),
	satisfiable_clauses(Normal_Clauses, Satisfiable_Clauses),
	format("abstracting, ", []), ttyflush,
	linear_clauses(Satisfiable_Clauses, Linear_Clauses),
	assert_linear_program(Linear_Clauses),
	concat(['../', File, '/', File, '.clpr'], LinearFile),
	pretty_print_program(Linear_Clauses, LinearFile),
	format("ordering... ", []), ttyflush,
	sccs_clauses(Linear_Clauses, Linear_SCCs),
	statistics(runtime, [_, Time]),
	format("~wms}", [Time]), ttyflush,
	true.

parse_program(File, Clauses) :-
	open(File, read, Stream),
	parse_clauses(Stream, Clauses),
	close(Stream).

parse_clauses(Stream, Cs) :-
   	read(Stream, Term),	
	expand_term(Term, Clause),
    	parse_clause(Clause, Stream, Cs).

parse_clause(end_of_file, _, []) :-
	!.
parse_clause((:- op(A, B, C)), Stream, Clauses) :- 
	!,
    	op(A, B, C),
	parse_clauses(Stream, Clauses).
parse_clause((:- _), Stream, Clauses) :- 
	!,
	parse_clauses(Stream, Clauses).
parse_clause((H :- B), Stream, Clauses) :- 
	!,
	Clauses = [(H :- B) | Rest_Clauses],
	parse_clauses(Stream, Rest_Clauses).
parse_clause(H, Stream, [(H :- true) | Clauses]) :-
	parse_clauses(Stream, Clauses).

satisfiable_clauses([], []).
satisfiable_clauses([Clause | Normal_Clauses], Satisfiable_Clauses) :-
	Clause = (_Head :- Body),
	my_not_memberchk(Body, false),
	!,
	Satisfiable_Clauses =  [Clause | Rest_Satisfiable_Clauses],
	satisfiable_clauses(Normal_Clauses, Rest_Satisfiable_Clauses).
satisfiable_clauses([_Clause | Normal_Clauses], Satisfiable_Clauses) :-
	satisfiable_clauses(Normal_Clauses, Satisfiable_Clauses).

normal_program(Clauses, Normal_Clauses) :-
	retractall(new_clause(_)),
	assert(new_clause(1)),
	normal_clauses(Clauses, Normal_Clauses, []).

normal_clauses([], Norm_Cs, Norm_Cs).
normal_clauses([C | Cs_In], Norm_Cs_Out, Norm_Cs_In) :-
	normal_clause(C, Cs_Out, Cs_In, Norm_Cs_Out, Norm_Cs_Middle),
	normal_clauses(Cs_Out, Norm_Cs_Middle, Norm_Cs_In).

normal_clause((H :- B), Cs_Out, Cs_In, [(R :- Gs) | Norm_Cs_In], Norm_Cs_In) :-
	H =.. [Name | HArgs],
	normal_args(HArgs, RArgs, Gs, Bs),
	R =.. [Name | RArgs],
    	normal_goal(B, Bs, Cs_Out, Cs_In).

normal_args(PArgs, RArgs, Gs_Out, Gs_In) :-
	normal_args(PArgs, RArgs, [], Gs_Out, Gs_In).

normal_args([], [], _, Gs, Gs).
normal_args([PArg | PArgs], [RArg | RArgs], OldArgs, Gs_Out, Gs_In) :-
	nonvar(PArg),
	!,
	Gs_Out = [(RArg = PArg) | Gs_Middle],
	normal_args(PArgs, RArgs, OldArgs, Gs_Middle, Gs_In).
normal_args([PArg | PArgs], [RArg | RArgs], OldArgs, Gs_Out, Gs_In) :-
	my_not_memberchk(OldArgs, PArg),
	!,
	RArg = PArg,
	normal_args(PArgs, RArgs, [PArg | OldArgs], Gs_Out, Gs_In).
normal_args([PArg | PArgs], [RArg | RArgs], OldArgs, [(PArg = RArg) | Gs_Out], Gs_In) :-
	normal_args(PArgs, RArgs, OldArgs, Gs_Out, Gs_In).

normal_goal(B, Bs, Cs_Out, Cs_In) :-
    	normal_goal(B, Bs, [], Cs_Out, Cs_In).

normal_goal(G, Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	var(G),
	!,
	normal_goal(call(G), Gs_Out, Gs_In, Cs_Out, Cs_In).
normal_goal((P; Q), Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	Gs_Out = [G | Gs_In], 
	term_variables((P; Q), Args),
	fresh_name(Name),
	G =.. [Name | Args],
	Cs_Out = [(G :- P), (G :- Q) | Cs_In].
normal_goal((P -> Q), Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	normal_goal((P, Q), Gs_Out, Gs_In, Cs_Out, Cs_In).
normal_goal(\+(G), Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	normal_goal(not(G), Gs_Out, Gs_In, Cs_Out, Cs_In).
normal_goal((P, Q), Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	normal_goal(P, Gs_Out, Gs_Middle, Cs_Out, Cs_Middle),
	normal_goal(Q, Gs_Middle, Gs_In, Cs_Middle, Cs_In).
normal_goal(!, Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	Gs_Out = Gs_In,
	Cs_Out = Cs_In.
normal_goal(not(G), Gs_Out, Gs_In, Cs_Out, Cs_In) :- 
	!,
	normal_goal(((G, !, fail); true), Gs_Out, Gs_In, Cs_Out, Cs_In).
normal_goal(G, Gs_Out, Gs_In, Cs_Out, Cs_In) :-
	predicate_property(G, built_in),
	!,
	Cs_Out = Cs_In,
	(normal_builtin(G, Gs_Out, Gs_In) ->
		true;
        	unsafe_to_ignore_normal_builtin(G, Gs_Out, Gs_In)). 
normal_goal(P, Gs_Out, Gs_In, Cs, Cs) :-
	P =.. [Name | PArgs],
	normal_args(PArgs, RArgs, Gs_Out, [R | Gs_In]),
	R =.. [Name | RArgs].

fresh_name(Fresh_Name)  :-
	retract(new_clause(Fresh_Number)),
	Fresh_Number1 is Fresh_Number + 1,
	asserta(new_clause(Fresh_Number1)),
	name(Fresh_Number, Number_String),
	name(Fresh_Name, [95, 95 | Number_String]).

normal_builtin((Term1 > Term2), Gs_Out, Gs_In) :-
	normal_builtin((Term2 < Term1), Gs_Out, Gs_In).
normal_builtin((Term1 < Term2), Gs_Out, Gs_In) :-
	Gs_Out = [(Term1 < Term2) | Gs_In].
normal_builtin((Term1 =< Term2), Gs_Out, Gs_In) :-
	Gs_Out = [(Term1 =< Term2) | Gs_In].
normal_builtin((Term1 >= Term2), Gs_Out, Gs_In) :-
	normal_builtin((Term2 =< Term1), Gs_Out, Gs_In).
normal_builtin((Term1 =:= Term2), Gs_Out, Gs_In) :-
	Gs_Out = [(Term1 =:= Term2) | Gs_In].
normal_builtin((Term1 =\= Term2), Gs_Out, Gs_In) :-
	Gs_Out = [(Term1 =\= Term2) | Gs_In].
normal_builtin((Term1 = Term2), Gs_Out, Gs_In) :-
	Gs_Out = [(Term1 = Term2) | Gs_In].
normal_builtin((Term1 =.. Term2), Gs_Out, Gs_In) :-
	Gs_Out = [(Term1 =.. Term2), compound(Term2) | Gs_In].
normal_builtin((Term1 == Term2), Gs_Out, Gs_In) :-
	normal_builtin((Term1 = Term2), Gs_Out, Gs_In).
normal_builtin((Term1 @< Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin((Term1 @< Term2), Gs_Out, Gs_In).
normal_builtin((Term1 @> Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin((Term1 @> Term2) , Gs_Out, Gs_In).
normal_builtin((Term1 @=< Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin((Term1 @=< Term2) , Gs_Out, Gs_In).
normal_builtin((Term1 @>= Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin((Term1 @>= Term2), Gs_Out, Gs_In).
normal_builtin((Term1 \== Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin((Term1 \== Term2), Gs_Out, Gs_In).
normal_builtin(abolish(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(abolish(Term1, Term2), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term1), integer(Term2) | Gs_In].
normal_builtin(abort, Gs_Out, Gs_In) :-
	normal_builtin(false, Gs_Out, Gs_In).
normal_builtin(absolute_file_name(Term1, Term2), Gs_Out, Gs_In) :- 
	Gs_Out = [atom(Term1), atom(Term2) | Gs_In].
normal_builtin(arg(Index, Term, Arg), Gs_Out, Gs_In) :-
	Gs_Out = [arg(Index, Term, Arg), integer(Index), compound(Term) | Gs_In].
normal_builtin(assert(Clause), Gs_Out, Gs_In) :-
        unsafe_to_ignore_normal_builtin(assert(Clause), Gs_Out, Gs_In). 
normal_builtin(asserta(Clause), Gs_Out, Gs_In) :-
        unsafe_to_ignore_normal_builtin(asserta(Clause), Gs_Out, Gs_In). 
normal_builtin(assertz(Clause), Gs_Out, Gs_In) :-
        unsafe_to_ignore_normal_builtin(assertz(Clause), Gs_Out, Gs_In). 
normal_builtin(atom(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(atomic(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atomic(Term) | Gs_In].
normal_builtin(atom_chars(Term1, Term2), Gs_Out, Gs_In) :- 
	Gs_Out = [atom(Term1), ground(Term2) | Gs_In].
normal_builtin(bagof(Target, G, Bag), Gs_Out, Gs_In) :- 
        safe_to_ignore_normal_builtin(bagof(Target, G, Bag), Gs_Out, Gs_In). 
normal_builtin(break, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(break, Gs_Out, Gs_In).
normal_builtin('C'(Term1, Term2, Term3), Gs_Out, Gs_In) :-
	normal_builtin((Term1 = [Term2 | Term3]), Gs_Out, Gs_In).
normal_builtin(call(G), Gs_Out, Gs_In) :- 
        safe_to_ignore_normal_builtin(call(G), Gs_Out, Gs_In). 
normal_builtin(callable(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(callable(Term), Gs_Out, Gs_In).
normal_builtin(character_count(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), integer(Term) | Gs_In].
normal_builtin(close(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(copy_term(Term1, Term2), Gs_Out, Gs_In) :-
        safe_to_ignore_normal_builtin(copy_term(Term1, Term2), Gs_Out, Gs_In). 
normal_builtin(compare(Term, _, _), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(compile(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(compound(Term), Gs_Out, Gs_In) :-
	Gs_Out = [compound(Term) | Gs_In].
normal_builtin(consult(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(current_atom(Term), Gs_Out, Gs_In) :-
        safe_to_ignore_normal_builtin(current_atom(Term), Gs_Out, Gs_In). 
normal_builtin(current_input(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(current_module(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(current_module(Term1, Term2), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term1), ground(Term2) | Gs_In].
normal_builtin(current_op(Term1, Term2, Term3), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term1), atom(Term2), atom(Term3) | Gs_In].
normal_builtin(current_output(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(current_predicate(Term1, Term2), Gs_Out, Gs_In) :-
        safe_to_ignore_normal_builtin(current_predicate(Term1, Term2), Gs_Out, Gs_In). 
normal_builtin(current_stream(Term1, Term2, Term3), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term1), atom(Term2), ground(Term3) | Gs_In].
normal_builtin(debug, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(debug, Gs_Out, Gs_In).
normal_builtin(debugging, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(debugging, Gs_Out, Gs_In).
normal_builtin(dif(Term1, Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(dif(Term1, Term2), Gs_Out, Gs_In).
normal_builtin(display(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(display(Term), Gs_Out, Gs_In).
normal_builtin(ensure_loaded(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(erase(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(expand_term(Term1, Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(expand_term(Term1, Term2), Gs_Out, Gs_In).
normal_builtin(fail, Gs_Out, Gs_In) :-
	normal_builtin(false, Gs_Out, Gs_In).
normal_builtin(false, Gs_Out, Gs_In) :-
	Gs_Out = [false | Gs_In].
normal_builtin(fileerrors, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(fileerrors, Gs_Out, Gs_In).
normal_builtin(findall(Target, G, Bag), Gs_Out, Gs_In) :- 
	safe_to_ignore_normal_builtin(findall(Target, G, Bag), Gs_Out, Gs_In).
normal_builtin(float(Term), Gs_Out, Gs_In) :-
	Gs_Out = [float(Term) | Gs_In].
normal_builtin(flush_output(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(format(Term, _), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(format(Stream, Term, _), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), ground(Term) | Gs_In].
normal_builtin(functor(_, Term1, Term2), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term1), integer(Term2) | Gs_In].
normal_builtin(garbage_collect, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(garbage_collect, Gs_Out, Gs_In).
normal_builtin(get(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(get(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), integer(Term) | Gs_In].
normal_builtin(get_char(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(get_char(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), atom(Term) | Gs_In].
normal_builtin(get0(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(get0(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), integer(Term) | Gs_In].
normal_builtin(halt, Gs_Out, Gs_In) :-
	normal_builtin(false, Gs_Out, Gs_In).
normal_builtin(halt(_Term), Gs_Out, Gs_In) :-
	normal_builtin(false, Gs_Out, Gs_In).
normal_builtin(gc, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(gc, Gs_Out, Gs_In).
normal_builtin(ground(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(help, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(help, Gs_Out, Gs_In).
normal_builtin(integer(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(instance(Term1, Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(instance(Term1, Term2), Gs_Out, Gs_In).
normal_builtin((Term1 is Term2), Gs_Out, Gs_In) :-
	Gs_Out = [(Term1 is Term2) | Gs_In].
normal_builtin(keysort(Term1, Term2), Gs_Out, Gs_In) :-
	Gs_Out = [keysort(Term1, Term2) | Gs_In].
normal_builtin(leash(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(length(Term1, Term2), Gs_Out, Gs_In) :-
	Gs_Out = [length(Term1, Term2), integer(Term2) | Gs_In].
normal_builtin(line_count(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), integer(Term) | Gs_In].
normal_builtin(line_position(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), integer(Term) | Gs_In].
normal_builtin(listing, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(listing, Gs_Out, Gs_In).
normal_builtin(listing(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(listing(Term), Gs_Out, Gs_In).
normal_builtin(load(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(name(Term1, Term2), Gs_Out, Gs_In) :-
	Gs_Out = [atomic(Term1), ground(Term2) | Gs_In].
normal_builtin(nl, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(nl, Gs_Out, Gs_In).
normal_builtin(nl(Stream), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream) | Gs_In].
normal_builtin(nodebug, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(nodebug, Gs_Out, Gs_In).
normal_builtin(nofileerrors, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(nofileerrors, Gs_Out, Gs_In).
normal_builtin(nogc, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(nogc, Gs_Out, Gs_In).
normal_builtin(nonvar(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(nonvar(Term), Gs_Out, Gs_In).
normal_builtin(nospyall, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(nospyall, Gs_Out, Gs_In).
normal_builtin(notrace, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(notrace, Gs_Out, Gs_In).
normal_builtin(number(Term), Gs_Out, Gs_In) :-
	Gs_Out = [number(Term) | Gs_In].
normal_builtin(number_chars(Term1, Term2), Gs_Out, Gs_In) :-
	Gs_Out = [number(Term1), ground(Term2) | Gs_In].
normal_builtin(numbervars(Term1, Term2, Term3), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term1), ground(Term2), ground(Term3) | Gs_In].
normal_builtin(op(Term1, Term2, Term3), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term1), atom(Term2), atom(Term3) | Gs_In].
normal_builtin(open(Term1, Term2, Stream), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term1), atom(Term2), ground(Stream) | Gs_In].
normal_builtin(open(Term1, Term2, Stream, Term3), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term1), atom(Term2), ground(Stream), atom(Term3) | Gs_In].
normal_builtin(open_null_stream(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(otherwise, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(otherwise, Gs_Out, Gs_In).
normal_builtin(peek_char(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(peek_char(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), integer(Term) | Gs_In].
normal_builtin(phrase(Term1, Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(phrase(Term1, Term2), Gs_Out, Gs_In).
normal_builtin(phrase(Term1, Term2, Term3), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(pharse(Term1, Term2, Term3), Gs_Out, Gs_In).
normal_builtin(portray_clause(Stream, _), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream) | Gs_In].
normal_builtin(predicate_property(_, Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(print(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(print(Term), Gs_Out, Gs_In).
normal_builtin(print(Stream, _), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream) | Gs_In].
normal_builtin(prolog_flag(Term1, Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(prolog_flag(Term1, Term2), Gs_Out, Gs_In).
normal_builtin(prolog_flag(Term1, Term2, Term3), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(prolog_flag(Term1, Term2, Term3), Gs_Out, Gs_In).
normal_builtin(prompt(Term1, Term2), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term1), atom(Term2) | Gs_In].
normal_builtin(put(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(put(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), integer(Term) | Gs_In].
normal_builtin(read(Stream, _), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream) | Gs_In].
normal_builtin(reconsult(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(read(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(read(Term), Gs_Out, Gs_In).
normal_builtin(repeat, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(repeat, Gs_Out, Gs_In).
normal_builtin(retract(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(retract(Term), Gs_Out, Gs_In).
normal_builtin(retractall(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(retractall(Term), Gs_Out, Gs_In).
normal_builtin(setof(Target, G, Bag), Gs_Out, Gs_In) :- 
        safe_to_ignore_normal_builtin(setof(Target, G, Bag), Gs_Out, Gs_In). 
normal_builtin(see(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(seen, Gs_Out, Gs_In):-
	safe_to_ignore_normal_builtin(seen, Gs_Out, Gs_In).
normal_builtin(seeing(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(set_input(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(set_output(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(simple(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(simple(Term), Gs_Out, Gs_In).
normal_builtin(skip(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(skip(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), integer(Term) | Gs_In].
normal_builtin(skip_line, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(skip_line, Gs_Out, Gs_In).
normal_builtin(skip_line(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(sort(Term1, Term2), Gs_Out, Gs_In) :-
	Gs_Out = [sort(Term1, Term2) | Gs_In].
normal_builtin(source_file(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(source_file(_, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(statistics, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(statistics, Gs_Out, Gs_In).
normal_builtin(statistics(Term1, Term2), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term1), ground(Term2) | Gs_In].
normal_builtin(stream_code(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), integer(Term) | Gs_In].
normal_builtin(system(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(system(Term), Gs_Out, Gs_In).
normal_builtin(subsumes_chk(Term1, Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(subsumes_chk(Term1, Term2), Gs_Out, Gs_In).
normal_builtin(tab(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(tab(Stream, Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream), integer(Term) | Gs_In].
normal_builtin(tell(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(telling(Term), Gs_Out, Gs_In) :-
	Gs_Out = [atom(Term) | Gs_In].
normal_builtin(told, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(told, Gs_Out, Gs_In).
normal_builtin(trace, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(trace, Gs_Out, Gs_In).
normal_builtin(true, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(true, Gs_Out, Gs_In).
normal_builtin(ttyflush, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(ttyflush, Gs_Out, Gs_In).
normal_builtin(ttyget(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(ttyget0(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(ttynl, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(ttynl, Gs_Out, Gs_In).
normal_builtin(ttyput(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(ttytab(Term), Gs_Out, Gs_In) :-
	Gs_Out = [integer(Term) | Gs_In].
normal_builtin(use_module(Term1, Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(use_module(Term1, Term2), Gs_Out, Gs_In).
normal_builtin(use_module(Term1, Term2, Term3), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(use_module(Term1, Term2, Term3), Gs_Out, Gs_In).
normal_builtin(var(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(var(Term), Gs_Out, Gs_In).
normal_builtin(version, Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(version, Gs_Out, Gs_In).
normal_builtin(version(Term), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Term) | Gs_In].
normal_builtin(write(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(write(Term), Gs_Out, Gs_In).
normal_builtin(write(Stream, _), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream) | Gs_In].
normal_builtin(write_canonical(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(write_canonical(Term), Gs_Out, Gs_In).
normal_builtin(write_canonical(Stream, _), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream) | Gs_In].
normal_builtin(writeq(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_normal_builtin(writeq(Term), Gs_Out, Gs_In).
normal_builtin(writeq(Stream, _), Gs_Out, Gs_In) :-
	Gs_Out = [ground(Stream) | Gs_In].

unsafe_to_ignore_normal_builtin(Atom, Gs, Gs) :-
	functor(Atom, Name, Arity),
	format("~n{warning, ignoring builtin ~w/~w...}", [Name, Arity]).

safe_to_ignore_normal_builtin(_Atom, Gs, Gs).

%%% LINEAR CONSTRAINTS

linear_clauses(Clauses, Linear_Clauses) :-
	empty_assoc(Assoc),
	linear_clauses(Clauses, Assoc, 1, Linear_Clauses).

linear_clauses([], _, _, []).
linear_clauses([Clause | Clauses], Assoc_In, Index, Linear_Clauses) :-
	linear_clause(Clause, Linear_Clause),
	(get_clause(Linear_Clause, Assoc_In) ->
		(Assoc_Out = Assoc_In,
		Linear_Clauses = Rest_Linear_Clauses);
		(put_clause(Linear_Clause, Assoc_In, Assoc_Out),
		Linear_Clauses = [Index-Linear_Clause | Rest_Linear_Clauses])),
	Index1 is Index + 1,
	linear_clauses(Clauses, Assoc_Out, Index1, Rest_Linear_Clauses).

get_clause(Clause, Assoc) :-
	copy_term(Clause, Copy_Clause),
	numbervars(Copy_Clause, 1, _),
	get_assoc(Copy_Clause, Assoc, _).

put_clause(Clause, Assoc_In, Assoc_Out) :-
	copy_term(Clause, Copy_Clause),
	numbervars(Copy_Clause, 1, _),
	put_assoc(Copy_Clause, Assoc_In, _, Assoc_Out).

linear_clause((H :- B), (H :- Linear_Gs_Out)) :-
	term_variables((H :- B), Vars),
	non_negativity(Vars, Linear_Gs_In),
	linear_goals(B, Linear_Gs_Out, Linear_Gs_In).
%	sort(Linear_Gs_Out, Sorted_Gs_Out).

non_negativity([], []).
non_negativity([Var | Vars], [scalar([Var-(-1)], =<, -1) | Gs]) :-		% FIXED FOR TERMSIZE 
	non_negativity(Vars, Gs).
	
linear_goals([], Linear_Gs, Linear_Gs).
linear_goals([G | Gs], Linear_Gs_Out, Linear_Gs_In) :-
	linear_goal(G, Linear_Gs_Out, Linear_Gs_Middle),
	linear_goals(Gs, Linear_Gs_Middle, Linear_Gs_In).

linear_goal(G, Gs_Out, Gs_In) :-
	predicate_property(G, built_in),
	!,
	term_size_builtin(G, Gs_Out, Gs_In).
linear_goal(Atom, [Atom | Gs_In], Gs_In).

term_size_builtin((Term1 < Term2), Gs_Out, Gs_In) :-				% checked
	safe_to_ignore_linear_builtin((Term1 =< Term2), Gs_Out, Gs_In).
term_size_builtin((Term1 =< Term2), Gs_Out, Gs_In) :-				% checked
	safe_to_ignore_linear_builtin((Term1 =< Term2), Gs_Out, Gs_In).
term_size_builtin((Term1 =:= Term2), Gs_Out, Gs_In) :-				% checked
	safe_to_ignore_linear_builtin((Term1 =:= Term2), Gs_Out, Gs_In).
term_size_builtin((Term1 =\= Term2), Gs_Out, Gs_In) :-				% checked
	safe_to_ignore_linear_builtin((Term1 =\= Term2), Gs_Out, Gs_In).
term_size_builtin((Term1 = Term2), [Scalar | Gs], Gs) :-			% checked
	term_sizes_equal(Term1, Term2, Scalar).
term_size_builtin((Term1 =.. Term2), [scalar(Sum, =<, NewValue) | Gs], Gs) :-	% checked
	term_sizes_equal(Term1, Term2, scalar(Sum, =, Value)),
	NewValue is Value - 2.
term_size_builtin(arg(_Index, Term, Arg), [scalar(Sum, =<, NewValue) | Gs], Gs) :-	% checked
	term_sizes_equal(Arg, Term, scalar(Sum, =, Value)),
	NewValue is Value - 1.
term_size_builtin(atom(Term), [Scalar | Gs], Gs) :-				% checked
	term_size_atomic(Term, Scalar).
term_size_builtin(atomic(Term), [Scalar | Gs], Gs) :-				% checked
	term_size_atomic(Term, Scalar).
term_size_builtin(compound(Term), [Scalar | Gs], Gs) :-				% checked
	term_size_compound(Term, Scalar).
term_size_builtin(float(Term), [Scalar | Gs], Gs) :-				% checked
	term_size_atomic(Term, Scalar).
term_size_builtin(ground(Term), Gs_Out, Gs_In) :-				% checked
	safe_to_ignore_linear_builtin(ground(Term), Gs_Out, Gs_In).
term_size_builtin(integer(Term), [Scalar | Gs], Gs) :-				% checked
	term_size_atomic(Term, Scalar).
term_size_builtin((Term1 is _Term2), [Scalar | Gs], Gs) :-			% checked
	term_size_atomic(Term1, Scalar).
term_size_builtin(keysort(Term1, Term2), [Scalar | Gs], Gs) :-			% checked
	term_sizes_equal(Term1, Term2, Scalar).
term_size_builtin(length(Term1, Term2), Gs_Out, Gs_In) :-			% checked
	safe_to_ignore_linear_builtin(length(Term1, Term2), Gs_Out, Gs_In).
term_size_builtin(number(Term), [Scalar | Gs], Gs) :-				% checked
	term_size_atomic(Term, Scalar).
term_size_builtin(sort(Term1, Term2), [scalar(Sum, =<, Value) | Gs], Gs) :-	% checked
	term_sizes_equal(Term2, Term1, scalar(Sum, =, Value)).

term_size_atomic(Term, scalar(Sum, =, NewValue)) :-
	term_size(Term, Sum, Value),
	NewValue is 1 - Value.							% FIXED FOR TERMSIZE

term_size_compound(Term, scalar(NegativeSum, =<, NewValue)) :-
	term_size(Term, Sum, Value),
	NewValue is Value,							% FIXED FOR TERMSIZE
	negative_sum(Sum, NegativeSum).

term_sizes_equal(Term1, Term2, scalar(Sum, =, Value)) :-
	term_size(Term1, Sum1, Value1),
	term_size(Term2, Sum2, Value2),
	Value is Value2 - Value1,
	negative_sum(Sum2, NegativeSum2),
	merge_sum(Sum1, NegativeSum2, Sum).

term_size(Term, Sum, Value) :-
	term_size(Term, [], 0, Sum, Value).

term_size(Term, SumAcc, ValueAcc, Sum, Value) :-
	var(Term),
	!,
	Value = ValueAcc,
	merge_sum([Term-1], SumAcc, Sum).
term_size(Term, SumAcc, ValueAcc, Sum, Value) :-
	atomic(Term),
	!,
	Sum = SumAcc,
	Value is ValueAcc + 1.
term_size(Term, SumAcc, ValueAcc, Sum, Value) :-
	NewValueAcc is ValueAcc + 1,							% FIXED FOR TERMSIZE
	Term =.. [_Name | Terms],
	term_sizes(Terms, SumAcc, NewValueAcc, Sum, Value).

term_sizes([], Sum, Value, Sum, Value).
term_sizes([Term | Terms], SumAcc, ValueAcc, Sum, Value) :-
	term_size(Term, SumAcc, ValueAcc, NewSumAcc, NewValueAcc),
	term_sizes(Terms, NewSumAcc, NewValueAcc, Sum, Value).

list_length_builtin((Term1 < Term2), Gs_Out, Gs_In) :-					
	safe_to_ignore_linear_builtin((Term1 =< Term2), Gs_Out, Gs_In).
list_length_builtin((Term1 =< Term2), Gs_Out, Gs_In) :-					
	safe_to_ignore_linear_builtin((Term1 =< Term2), Gs_Out, Gs_In).
list_length_builtin((Term1 =:= Term2), Gs_Out, Gs_In) :-					
	safe_to_ignore_linear_builtin((Term1 =:= Term2), Gs_Out, Gs_In).
list_length_builtin((Term1 =\= Term2), Gs_Out, Gs_In) :-				
	safe_to_ignore_linear_builtin((Term1 =\= Term2), Gs_Out, Gs_In).
list_length_builtin((Term1 = Term2), [Scalar | Gs], Gs) :-				
	list_lengths_equal(Term1, Term2, Scalar).
list_length_builtin((Term1 =.. Term2), Gs_Out, Gs_In) :-		
	safe_to_ignore_linear_builtin((Term1 =.. Term2), Gs_Out, Gs_In).
list_length_builtin(arg(Index, Term, Arg), Gs_Out, Gs_In) :-
	safe_to_ignore_linear_builtin(arg(Index, Term, Arg), Gs_Out, Gs_In).
list_length_builtin(atom(Term), Gs_Out, Gs_In) :-					
	safe_to_ignore_linear_builtin(atom(Term), Gs_Out, Gs_In).
list_length_builtin(atomic(Term), Gs_Out, Gs_In) :-					
	safe_to_ignore_linear_builtin(atomic(Term), Gs_Out, Gs_In).
list_length_builtin(compound(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_linear_builtin(compound(Term), Gs_Out, Gs_In).
list_length_builtin(float(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_linear_builtin(float(Term), Gs_Out, Gs_In).
list_length_builtin(ground(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_linear_builtin(ground(Term), Gs_Out, Gs_In).
list_length_builtin(integer(Term), Gs_Out, Gs_In) :-
	safe_to_ignore_linear_builtin(integer(Term), Gs_Out, Gs_In).
list_length_builtin((Term1 is Term2), Gs_Out, Gs_In) :-
	safe_to_ignore_linear_builtin((Term1 is Term2), Gs_Out, Gs_In).
list_length_builtin(keysort(Term1, Term2), [Scalar | Gs], Gs) :-		
	list_lengths_equal(Term1, Term2, Scalar).
list_length_builtin(length(Term1, Term2), Gs_Out, Gs_In) :-			
	safe_to_ignore_linear_builtin(length(Term1, Term2), Gs_Out, Gs_In).
list_length_builtin(number(Term), Gs_Out, Gs_In) :-	
	safe_to_ignore_linear_builtin(number(Term), Gs_Out, Gs_In).
list_length_builtin(sort(Term1, Term2), [scalar(Sum, =<, Value) | Gs], Gs) :-
	list_lengths_equal(Term2, Term1, scalar(Sum, =, Value)).

list_lengths_equal(Term1, Term2, scalar(Sum, =, Value)) :-
	list_length_term(Term1, Sum1, Value1),
	list_length_term(Term2, Sum2, Value2),
	Value is Value2 - Value1,
	negative_sum(Sum2, NegativeSum2),
	merge_sum(Sum1, NegativeSum2, Sum).

list_length_term(Term, Sum, Value) :-
	list_length_term(Term, [], 0, Sum, Value).

list_length_term(Term, SumAcc, ValueAcc, Sum, Value) :-
	var(Term),
	!,
	Value = ValueAcc,
	merge_sum([Term-1], SumAcc, Sum).
list_length_term([], SumAcc, ValueAcc, Sum, Value) :-
	!,
	Value = ValueAcc,
	Sum = SumAcc.
list_length_term([_ | Term], SumAcc, ValueAcc, Sum, Value) :-
	!,
	NewValueAcc is ValueAcc + 1,
	list_length_term(Term, SumAcc, NewValueAcc, Sum, Value).
list_length_term(_, SumAcc, ValueAcc, Sum, Value) :-
	Value = ValueAcc,
	merge_sum([_-1], SumAcc, Sum).

safe_to_ignore_linear_builtin(_Atom, Gs, Gs).

unsafe_to_ignore_linear_builtin(Atom, Gs, Gs) :-
	functor(Atom, Name, Arity),
	format("~n{warning, ignoring builtin ~w...}", [Name, Arity]).

%%% GROUNDNESS CONSTRAINTS

ground_clauses(Clauses, Ground_Clauses) :-
	empty_assoc(Assoc),
	ground_clauses(Clauses, Assoc, Ground_Clauses).

ground_clauses([], _, []).
ground_clauses([Clause | Clauses], Assoc_In, Ground_Clauses) :-
	ground_clause(Clause, Ground_Clause),
	(get_clause(Ground_Clause, Assoc_In) ->
		(Assoc_Out = Assoc_In,
		Ground_Clauses = Rest_Ground_Clauses);
		(put_clause(Ground_Clause, Assoc_In, Assoc_Out),
		Ground_Clauses = [Ground_Clause | Rest_Ground_Clauses])),
	ground_clauses(Clauses, Assoc_Out, Rest_Ground_Clauses).

ground_clause((H :- Body), (H :- Sorted_Body)) :-
	ground_goals(Body, Ground_Body, []),
	sort(Ground_Body, Sorted_Body).

ground_goals([], Ground_Gs, Ground_Gs).
ground_goals([G | Gs], Ground_Gs_Out, Ground_Gs_In) :-
	ground_goal(G, Ground_Gs_Out, Ground_Gs_Middle),
	ground_goals(Gs, Ground_Gs_Middle, Ground_Gs_In).

ground_goal(G, Gs_Out, Gs_In) :-
	predicate_property(G, built_in),
	!,
	ground_builtin(G, Gs_Out, Gs_In).
ground_goal(Atom, [Atom | Gs_In], Gs_In).

ground_builtin((Term1 < Term2), Gs_Out, Gs_In) :-					
	ground_term(Term1-Term2, Gs_Out, Gs_In).
ground_builtin((Term1 =< Term2), Gs_Out, Gs_In) :-					
	ground_term(Term1-Term2, Gs_Out, Gs_In).
ground_builtin((Term1 =:= Term2), Gs_Out, Gs_In) :-					
	ground_term(Term1-Term2, Gs_Out, Gs_In).
ground_builtin((Term1 =\= Term2), Gs_Out, Gs_In) :-					
	ground_term(Term1-Term2, Gs_Out, Gs_In).
ground_builtin((Term1 = Term2), Gs_Out, Gs_In) :-					
	ground_terms_equal(Term1, Term2, Gs_Out, Gs_In).
ground_builtin((Term1 =.. Term2), Gs_Out, Gs_In) :-					
	ground_terms_equal(Term1, Term2, Gs_Out, Gs_In).
ground_builtin(arg(Index, Term, Arg), Gs_Out, Gs_In) :-		
	ground_term(Index, Gs_Out, Gs_Middle),
	ground_terms_equal(Term, Arg-_, Gs_Middle, Gs_In).
ground_builtin(atom(Term), Gs_Out, Gs_In) :-					
	ground_term(Term, Gs_Out, Gs_In).
ground_builtin(atomic(Term), Gs_Out, Gs_In) :-					
	ground_term(Term, Gs_Out, Gs_In).
ground_builtin(compound(Term), Gs_Out, Gs_In) :-					
	safe_to_ignore_ground_builtin(compound(Term), Gs_Out, Gs_In).
ground_builtin(float(Term), Gs_Out, Gs_In) :-
	ground_term(Term, Gs_Out, Gs_In).
ground_builtin(ground(Term), Gs_Out, Gs_In) :-				
	ground_term(Term, Gs_Out, Gs_In).
ground_builtin(integer(Term), Gs_Out, Gs_In) :-					
	ground_term(Term, Gs_Out, Gs_In).
ground_builtin((Term1 is Term2), Gs_Out, Gs_In) :-					
	ground_term(Term1-Term2, Gs_Out, Gs_In).
ground_builtin(keysort(Term1, Term2), Gs_Out, Gs_In) :-			
	ground_terms_equal(Term1, Term2, Gs_Out, Gs_In).
ground_builtin(length(Term1, Term2), Gs_Out, Gs_In) :-				
	ground_term(Term1-Term2, Gs_Out, Gs_In).
ground_builtin(number(Term), Gs_Out, Gs_In) :-		
	ground_term(Term, Gs_Out, Gs_In).
ground_builtin(sort(Term1, Term2), Gs_Out, Gs_In) :-	
	ground_terms_equal(Term1, Term2, Gs_Out, Gs_In).

ground_term(Term, Gs_Out, Gs_In) :-
	term_variables(Term, Vars),
	ground_variables(Vars, Gs_Out, Gs_In).

ground_variables([], Gs, Gs).
ground_variables([Var | Vars], Gs_Out, Gs_In) :-
	ground_interface(Var, [], Gs_Out, Gs_Middle),
	ground_variables(Vars, Gs_Middle, Gs_In). 

ground_terms_equal(Term1, Term2, Gs_Out, Gs_In) :-
	ground_interface(Var, Vars1, Gs_Out, Gs_Middle),
	ground_interface(Var, Vars2, Gs_Middle, Gs_In),
	term_variables(Term1, Vars1),
	term_variables(Term2, Vars2).

ground_interface(Var, Vars, [ground(Var, Vars) | Gs], Gs).

safe_to_ignore_ground_builtin(_Atom, Gs, Gs).

unsafe_to_ignore_ground_builtin(Atom, Gs, Gs) :-
	functor(Atom, Name, Arity),
	format("~n{warning, ignoring builtin ~w/~w...}", [Name, Arity]).

%%% SCCS

sccs_clauses(LinearClauses, SimpleSCCs) :-
	empty_assoc(Empty_Assoc),
	names_to_indices(LinearClauses, Empty_Assoc, Assoc),
	clauses_to_edges(LinearClauses, Assoc, Edges),
	vertices_edges_to_ugraph([], Edges, Graph),
 	reduce(Graph, Reduce),
 	top_sort(Reduce, SCCs),
 	sccs_simplify(SCCs, SimpleSCCs).

names_to_indices([], Assoc, Assoc).
names_to_indices([Index-(Head :- _) | Clauses], Assoc_In, Assoc_Out) :-
	functor(Head, Name, Arity),
	get_assoc(Name/Arity, Assoc_In, Indices, Assoc_Middle, [Index | Indices]),
	!,
	names_to_indices(Clauses, Assoc_Middle, Assoc_Out).
names_to_indices([Index-(Head :- _) | Clauses], Assoc_In, Assoc_Out) :-
	functor(Head, Name, Arity),
	put_assoc(Name/Arity, Assoc_In, [Index], Assoc_Middle),
	names_to_indices(Clauses, Assoc_Middle, Assoc_Out).

clauses_to_edges([], _Assoc, []).
clauses_to_edges([Index-(_Head :- Body) | Clauses], Assoc, Edges_Out) :-
	goals_to_edges(Body, Index, Assoc, Edges_Out, Edges_In),
	clauses_to_edges(Clauses, Assoc, Edges_In).

goals_to_edges([], _Index, _Assoc, Edges, Edges).
goals_to_edges([scalar(_, _, _) | Atoms], Index, Assoc, Edges_Out, Edges_In) :-
	!,
	goals_to_edges(Atoms, Index, Assoc, Edges_Out, Edges_In).
%goals_to_edges([Atom | Atoms], Index, Assoc, Edges_Out, Edges_In) :-
%	predicate_property(Atom, built_in),
%	!,
%	goals_to_edges(Atoms, Index, Assoc, Edges_Out, Edges_In).
goals_to_edges([Atom | Atoms], Index, Assoc, Edges_Out, Edges_In) :-
	functor(Atom, Name, Arity),
	get_assoc(Name/Arity, Assoc, Indices),
	!,
	indices_to_edges(Indices, Index, Edges_Out, Edges_Middle),
	goals_to_edges(Atoms, Index, Assoc, Edges_Middle, Edges_In).
goals_to_edges([Atom | Atoms], Index, Assoc, Edges_Out, Edges_In) :-
	functor(Atom, Name, Arity),
	format("~n{warning, predicate ~w used but not defined}", [Name/Arity]),
	goals_to_edges(Atoms, Index, Assoc, Edges_Out, Edges_In).

indices_to_edges([], _Index, Edges, Edges).
indices_to_edges([OtherIndex | Indices], Index, [OtherIndex-Index | Edges_Out], Edges_In) :-
	indices_to_edges(Indices, Index, Edges_Out, Edges_In).

sccs_simplify([], []).
sccs_simplify([SCC | SCCs], [SimpleSCC | SimpleSCCs]) :-
	scc_simplify(SCC, SimpleSCC),
	sccs_simplify(SCCs, SimpleSCCs).

scc_simplify([Index], SimpleSCC) :-
	head_to_linear_clause(Index, Head, Body),
	\+ memberchk(Head, Body),
	!,
	SimpleSCC = Index.
scc_simplify(SCC, SCC).

%%% INFRASTRUCTURE

assert_normal_program(Clauses) :-
	retractall(head_to_normal_clause(_, _)),
	assert_normal_clauses(Clauses).

assert_linear_program(Clauses) :-
	retractall(head_to_linear_clause(_, _, _)),
	assert_linear_clauses(Clauses).

assert_ground_program(Clauses) :-
	retractall(head_to_ground_clause(_, _)),
	assert_ground_clauses(Clauses).

assert_normal_clauses([]).
assert_normal_clauses([(Head :- Body) | Clauses]) :-
	assert(head_to_normal_clause(Head, Body)),
	assert_normal_clauses(Clauses).

assert_linear_clauses([]).
assert_linear_clauses([Index-(Head :- Body) | Clauses]) :-
	assert(head_to_linear_clause(Index, Head, Body)),
	assert_linear_clauses(Clauses).

assert_ground_clauses([]).
assert_ground_clauses([(Head :- Body) | Clauses]) :-
	assert(head_to_ground_clause(Head, Body)),
	assert_ground_clauses(Clauses).

pretty_print_program(Clauses, File) :-
	open(File, write, Stream),
	pretty_print_clauses(Clauses, Stream),
	close(Stream).

pretty_print_clauses([], _).
pretty_print_clauses([Index-(Head :- Atoms) | Clauses], Stream) :-
	!,
	format(Stream, "% ~w~n", [Index]),
	scalars_to_goal(Atoms, Body),
	portray_clause(Stream, (Head :- Body)),
	pretty_print_clauses(Clauses, Stream).
pretty_print_clauses([(Head :- Atoms) | Clauses], Stream) :-
	scalars_to_goal(Atoms, Body),
	portray_clause(Stream, (Head :- Body)),
	pretty_print_clauses(Clauses, Stream).

