:- module(linear_fixpoint, [linear_lfp_program/2, linear_lfp_pretty_print/1]).
:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(polyhedra).
:- use_module(utilities).

:- dynamic linear_atom/3.
:- dynamic redo/0.

linear_lfp_program(SCCs, Trace) :-
	retractall(redo),
	retractall(linear_atom(_, _, _)),
	lfp_sccs(SCCs, Trace).

lfp_sccs([], _).
lfp_sccs([SCC | SCCs], Trace) :-
	format("~nSCC: ~w ", [SCC]), ttyflush,
	lfp_scc(SCC, Trace),
	lfp_sccs(SCCs, Trace).

lfp_scc(Index, Trace) :-
	integer(Index),
	!,
	lfp_scc_pass([Index], Trace, false).
lfp_scc(Indices, Trace) :-
	lfp_scc_pass(Indices, Trace, true),
	retract(redo),
	!,
	lfp_scc(Indices, Trace).
lfp_scc(_, _).
	
lfp_scc_pass([], _, _).
lfp_scc_pass([Index | Indices], Trace, Widen) :-
	format("~nclause: ~w ", [Index]), ttyflush,
	lfp_clause(Index, Trace, Widen),
	lfp_scc_pass(Indices, Trace, Widen).

lfp_clause(Index, Trace, Widen) :-
        abstract:head_to_linear_clause(Index, Atom, Atoms), 
	lfp_atoms(Atoms, Scalars),
	!,
	Atom =.. [_ | Args],
	project_lite(Scalars, Args, Project),
        cond_assert(Atom, Project, Trace, Widen).
lfp_clause(_Index, _Trace, _Widen).

lfp_atoms([], []).
lfp_atoms([Atom | Atoms], [Atom | Scalars]) :- 
	functor(Atom, scalar, 3),
	!,
	lfp_atoms(Atoms, Scalars).
lfp_atoms([Atom | Atoms], Scalars) :- 
	linear_atom(Atom, Scalars, RestScalars),
	lfp_atoms(Atoms, RestScalars).

redo_assert :-
	redo -> true; asserta(redo).

cond_assert(Atom, Scalars, Trace, Widen) :-
	linear_atom(Atom, FixScalars, []), 
	!,
	cond_assert_aux(Atom, Scalars, FixScalars, Trace, Widen).
cond_assert(Atom, Scalars, Trace, _Widen) :-
	stepper(Trace, new, Atom, Scalars),
	redo_assert,
	append(Scalars, Scalars_In, Scalars_Out),
	asserta(linear_atom(Atom, Scalars_Out, Scalars_In)). 

cond_assert_aux(Atom, Scalars, FixScalars, Trace, _Widen) :-
	entails(Scalars, FixScalars),
	!,
	stepper_aux(Trace, new_entails, Atom, Scalars),
	stepper(Trace, fix_entails, Atom, FixScalars).
cond_assert_aux(Atom, Scalars, FixScalars, Trace, _Widen) :-
	Atom =.. [Name | Args],
	stepper_aux(Trace, fix_lub, Atom, FixScalars),
	stepper_aux(Trace, new_lub, Atom, Scalars),
	convex_hull_lite(Args, Scalars, Args, FixScalars, LubArgs, LubScalars),
	LubAtom =.. [Name | LubArgs],
	LubAtom = Atom,
	stepper(Trace, result_lub, Atom, LubScalars),
	redo_assert,
	retract(linear_atom(Atom, _, _)),
	append(LubScalars, Scalars_In, Scalars_Out),
	asserta(linear_atom(Atom, Scalars_Out, Scalars_In)). 

stepper(off, _, _, _).
stepper(step, Message, Atom, Scalars) :-
	stepper_aux(log, Message, Atom, Scalars),
	read(Action),
	stepper_aux_aux(Action).
stepper(log, Message, Atom, Scalars) :-
	stepper_aux(log, Message, Atom, Scalars).

stepper_aux(off, _, _, _).
stepper_aux(step, Message, Atom, Scalars) :-
	stepper_aux(log, Message, Atom, Scalars).
stepper_aux(log, Message, Atom, Scalars) :-
	format("~n~w: ", [Message]), 
	scalars_to_goal(Scalars, Body),
	portray_clause((Atom :- Body)),
	ttyflush.

stepper_aux_aux(trace) :-
	!,
	trace.
stepper_aux_aux(abort) :-
	!,
	abort.
stepper_aux_aux(_).

%% PRETTY PRINTING CODE

linear_lfp_pretty_print(File) :-
	open(File, write, Stream),
	linear_lfp_pretty_print_aux(Stream).

linear_lfp_pretty_print_aux(Stream) :-
	linear_atom(Head, Scalars, []),
%	Head =.. [_ | Args],
%	reverse(Args, [Depth | RestArgs]),
%	depth_project(RestArgs, Depth, Scalars, Scalars, DepthScalars),
%	scalars_to_goal(DepthScalars, Body),
	scalars_to_goal(Scalars, Body),
	portray_clause(Stream, (Head :- Body)),
	fail.
linear_lfp_pretty_print_aux(Stream) :-
	close(Stream).

depth_project([], _, _, DepthScalars, DepthScalars).
depth_project([Arg | Args], Depth, Scalars, AccScalars, DepthScalars) :-
	project(Scalars, [Arg, Depth], Project),
	union(AccScalars, Project, NewAccScalars),
	depth_project(Args, Depth, Scalars, NewAccScalars, DepthScalars).

