%:- module(polyhedra, [convex_hull/6, project/3, projections/2, widen/3, 
%		merge_sum/3, negative_sum/2, multiply_sum/3,
%		scalars_to_goal/2, unsatisfiable/1,
%		entails/2, compress/2, test/2]).
:- use_module(library(assoc)).
:- use_module(library(terms)).
:- use_module(library(lists)).
:- use_module(utilities).

%%% SCRATCH AREA

widen([], _, []).
widen([Scalar2 | Scalars2], Scalars1, WidenScalars) :-
	single_entailed(Scalar2, Scalars1),
	!,
	WidenScalars = [Scalar2 | RestWidenScalars],
	widen(Scalars2, Scalars1, RestWidenScalars).
widen([_ | Scalars2], Scalars1, WidenScalars) :-
	widen(Scalars2, Scalars1, WidenScalars).

compress(Scalars, DenseScalars) :- compress(Scalars, 8, DenseScalars).
maximally_compress(Scalars, DenseScalars) :- compress(Scalars, 2, DenseScalars).

compress(Scalars, _, DenseScalars) :-
	unsatisfiable_lite(Scalars),
	!,
	DenseScalars = [false].
compress(Scalars, Threshold, DenseScalars) :-
	compress_lite(Scalars, [], DenseLiteScalars),
	length(DenseLiteScalars, Length),
	(Length >= Threshold ->
		compress_heavy(DenseLiteScalars, [], DenseScalars);
		DenseScalars = DenseLiteScalars).

compress_lite([], Scalars, Scalars).
compress_lite([Scalar | Scalars], Residual, DenseScalars) :-
	append(Scalars, Residual, RestScalars),
	single_entailed_lite(Scalar, RestScalars),
	!,
	compress_lite(Scalars, Residual, DenseScalars).
compress_lite([Scalar | Scalars], Residual, DenseScalars) :-
	compress_lite(Scalars, [Scalar | Residual], DenseScalars).

entails(Scalars1, Scalars2) :-
	entails_aux(Scalars2, Scalars1).

entails_aux([], _).
entails_aux([Scalar2 | Scalars2], Scalars1) :-
	single_entailed(Scalar2, Scalars1),
	entails_aux(Scalars2, Scalars1).

single_entailed(Scalar, Scalars) :-
	(single_entailed_lite(Scalar, Scalars) ->
		true;
		single_entailed_heavy(Scalar, Scalars)).
%  true == single_entailed is not invoked from compress 
	
single_entailed_lite(scalar([], =, Value2), _) :-
	!,
	Value2 == 0.
single_entailed_lite(scalar([], =<, Value2), _) :-
	!,
	0 =< Value2.
single_entailed_lite(scalar(Sum, =, Value2), Scalars1) :-
	find_scalar_single(Scalars1, Sum, =, Value1),
	!,
	Value1 == Value2.
single_entailed_lite(scalar(Sum, =<, Value2), Scalars1) :-
	find_scalar_many(Scalars1, Sum, _, Value1),
	Value1 =< Value2.

single_entailed_heavy(scalar(Sum, =<, Value), Scalars) :- 
	multiply_sum(Sum, -1, NegSum),
	OffsetValue is -(Value + 1),
	unsatisfiable([scalar(NegSum, =<, OffsetValue) | Scalars]).
single_entailed_heavy(scalar(Sum, =, Value), Scalars) :- 
	OffsetValue is Value - 1,	
	NegOffsetValue is -(Value + 1),
	multiply_sum(Sum, -1, NegSum),
	unsatisfiable([scalar(Sum, =<, OffsetValue) | Scalars]),
	unsatisfiable([scalar(NegSum, =<, NegOffsetValue) | Scalars]).

unsatisfiable(Scalars) :- 
	unsatisfiable_lite(Scalars),
	!.
unsatisfiable(Scalars) :- 
	copy_term(Scalars, Copy_Scalars),
	\+ post_scalars(Copy_Scalars).

sort_scalars([], []).
sort_scalars([scalar(Sum, RelOp, Value) | Scalars], [scalar(SortedSum, RelOp, Value) | SortedScalars]) :-
	keysort(Sum, SortedSum),
	sort_scalars(Scalars, SortedScalars).

post_scalars([]).
post_scalars([scalar(Sum, RelOp, Value) | Scalars]) :-
	scalars_to_expr(Sum, Expr),
	(RelOp == '=' -> {Expr = Value}; {Expr =< Value}),
	post_scalars(Scalars).

scalars_to_expr([], 0).
scalars_to_expr([X-Coeff | Sum], ((Coeff*X) + RestExpr)) :-
	scalars_to_expr(Sum, RestExpr).

unsatisfiable_lite([false | _]) :-
	!.
unsatisfiable_lite([scalar([], =, Value) | _]) :-
	Value \== 0,
	!.
unsatisfiable_lite([scalar([], =<, Value) | _]) :-
	Value < 0,
	!.
unsatisfiable_lite([_ | Scalars]) :-
	unsatisfiable_lite(Scalars).

find_scalar_single(Scalars, Sum, RelOp, FindValue) :-
	once(find_scalar_many(Scalars, Sum, RelOp, FindValue)).

find_scalar_many([scalar(Sum, RelOp, Value) | _], FindSum, RelOp, FindValue) :-
	Sum == FindSum, 
	FindValue = Value.
find_scalar_many([_ | Scalars], FindSum, RelOp, FindValue) :-
	find_scalar_many(Scalars, FindSum, RelOp, FindValue).

convex_hull(Xs, ScalarsXs, Ys, ScalarsYs, Zs, ScalarsZs) :- % sets Xs and Ys can overlap
	copy_term(Ys-ScalarsYs, CopyYs-CopyScalarsYs),
    	scale(ScalarsXs, Sigma1, [], Scalars1),
    	scale(CopyScalarsYs, Sigma2, Scalars1, Scalars2),
    	add_vect(Xs, CopyYs, Zs, Scalars2, Scalars3),
	Scalars4 = [scalar([Sigma1-(-1)], =<, 0),
		scalar([Sigma2-(-1)], =<, 0),
		scalar([Sigma1-1, Sigma2-1], =, 1) | Scalars3],
    	project(Scalars4, Zs, ScalarsZs).

project(Scalars, TargetVars, Project) :-
	compress(Scalars, DenseScalars),
	term_variables(DenseScalars, Vars),
	my_subtract(TargetVars, Vars, ResidueVars),
	gaussian_elim(ResidueVars, RemainVars, DenseScalars, PartProject),
	fourier_motzkin_elim(RemainVars, PartProject, Project).

gaussian_elim([], [], Project, Project).
gaussian_elim([EqVar | Vars], NEqVars, Scalars, Project) :-
	find_eq(Scalars, EqVar, EqSum, EqValue, Residue),
	!,
	substitute_eq(Residue, EqVar, EqSum, EqValue, PartProject),
	gaussian_elim(Vars, NEqVars, PartProject, Project).
gaussian_elim([Var | Vars], [Var | NEqVars], Scalars, Project) :-
	gaussian_elim(Vars, NEqVars, Scalars, Project).

fourier_motzkin_elim([], Project, Project).
fourier_motzkin_elim([Var | Vars], Scalars, Project) :-
	empty_assoc(Assoc_In),
	rank_vars(Scalars, [Var | Vars], Assoc_In, Assoc_Out),
	assoc_to_list(Assoc_Out, List),
	search_vars(List, BestVar, RestVars),
	find_ineqs(Scalars, BestVar, Plus, Minus, Residue),
	product_ineqs_ineqs(Plus, Minus, [], Pairs),
	keysort(Pairs, OrderedPairs),
	prefix_ineqs_ineqs(OrderedPairs, 64, PrefixPairs),		% CRITICAL PARAMETER
	substitute_ineqs_ineqs(PrefixPairs, BestVar, PartProject, Residue),
	compress(PartProject, DenseProject),
	(DenseProject == [false] ->
		Project = [false];
		fourier_motzkin_elim(RestVars, DenseProject, Project)).

prefix_ineqs_ineqs([], _, []).
prefix_ineqs_ineqs([_-Pair|RestPairs], Index, Prefix) :-
	Index1 is Index - 1,
	(Index == 0 ->
		Prefix = [];
		(Prefix = [Pair|RestPrefix],
		prefix_ineqs_ineqs(RestPairs, Index1, RestPrefix))).

product_ineqs_ineqs([], _Minus, Pairs, Pairs).
product_ineqs_ineqs([Scalar | Plus], Minus, Pairs_In, Pairs_Out) :-
	product_ineqs_ineq(Minus, Scalar, Pairs_In, Pairs_Middle),
	product_ineqs_ineqs(Plus, Minus, Pairs_Middle, Pairs_Out).

product_ineqs_ineq([], _Scalar, Pairs, Pairs).
product_ineqs_ineq([Neg_Scalar | Minus], Scalar, Pairs_In, Pairs_Out) :-
	Scalar = scalar(Sum, _, _),
	Neg_Scalar = scalar(Neg_Sum, _, _),
	length(Sum, Pos_Size),
	length(Neg_Sum, Neg_Size),
	Size is Pos_Size + Neg_Size - 2,
	Pairs_Middle = [Size-(Scalar-Neg_Scalar) | Pairs_In],
	product_ineqs_ineq(Minus, Scalar, Pairs_Middle, Pairs_Out).

rank_vars([], _, Assoc, Assoc).
rank_vars([scalar(Sum, _, _) | Scalars], Vars, Assoc_In, Assoc_Out) :-
	rank_vars_aux(Sum, Vars, Assoc_In, Assoc_Middle),
	rank_vars(Scalars, Vars, Assoc_Middle, Assoc_Out).

rank_vars_aux([], _, Assoc, Assoc).
rank_vars_aux([Var-_Coeff | Sum], Vars, Assoc_In, Assoc_Out) :-
	my_not_memberchk(Vars, Var),
	!,
	rank_vars_aux(Sum, Vars, Assoc_In, Assoc_Out).
rank_vars_aux([Var-Coeff | Sum], Vars, Assoc_In, Assoc_Out) :-
	get_assoc(Var, Assoc_In, (Plus,Minus), Assoc_Middle, (NewPlus,NewMinus)),
	!,
	(Coeff > 0 -> 
		(NewPlus is Plus + 1, NewMinus = Minus);
		(NewMinus is Minus + 1, NewPlus = Plus)),
	rank_vars_aux(Sum, Vars, Assoc_Middle, Assoc_Out).
rank_vars_aux([Var-Coeff | Sum], Vars, Assoc_In, Assoc_Out) :-
	(Coeff > 0 -> 
		put_assoc(Var, Assoc_In, (1,0), Assoc_Middle);
		put_assoc(Var, Assoc_In, (0,1), Assoc_Middle)),
	rank_vars_aux(Sum, Vars, Assoc_Middle, Assoc_Out).

search_vars([Var-(Plus,Minus) | Pairs], BestVar, RestVars) :-
	Rank is Plus * Minus,
	search_vars(Pairs, Rank, Var, [], BestVar, RestVars).

search_vars([], _, Var, Vars, Var, Vars).
search_vars([Var-(Plus,Minus) | Pairs], AccRank, AccVar, AccVars, BestVar, RestVars) :-
	Rank is Plus * Minus,
	Rank < AccRank,
	!,
	search_vars(Pairs, Rank, Var, [AccVar | AccVars], BestVar, RestVars).
search_vars([Var-_ | Pairs], AccRank, AccVar, AccVars, BestVar, RestVars) :-
	search_vars(Pairs, AccRank, AccVar, [Var | AccVars], BestVar, RestVars).

find_eq([scalar(Sum, =, Value) | Scalars], EqVar, EqSum, EqValue, Residue) :-
	sum_member(Sum, EqVar),
	!,
	EqSum = Sum,
	EqValue = Value,
	Residue = Scalars.
find_eq([Scalar | Scalars], EqVar, EqSum, EqValue, [Scalar | Residue]) :-
	find_eq(Scalars, EqVar, EqSum, EqValue, Residue).

find_ineqs([], _, [], [], []).
find_ineqs([Scalar | Scalars], Var, NewPlus, NewMinus, Residue) :-
	Scalar = scalar(Sum, =<, _Value),
	retrieve_coeff(Sum, Var, Coeff), 
	!,
	(Coeff > 0 ->
		(NewPlus = [Scalar | Plus], NewMinus = Minus);
		(NewPlus = Plus, NewMinus = [Scalar | Minus])),
	find_ineqs(Scalars, Var, Plus, Minus, Residue).
find_ineqs([Scalar | Scalars], Var, Plus, Minus, [Scalar | Residue]) :-
	find_ineqs(Scalars, Var, Plus, Minus, Residue).

substitute_eq([], _, _, _, []).
substitute_eq([scalar(Sum, RelOp, Value) | Scalars], EqVar, EqSum, EqValue, Project) :-
	sum_member(Sum, EqVar),
	!,
	single_substitute_eq(Sum, RelOp, Value, EqVar, EqSum, EqValue, Project, RestProject),
	substitute_eq(Scalars, EqVar, EqSum, EqValue, RestProject).
substitute_eq([Scalar | Scalars], EqVar, EqSum, EqValue, [Scalar | Project]) :-
	substitute_eq(Scalars, EqVar, EqSum, EqValue, Project).

single_substitute_eq(Sum, RelOp, Value, EqVar, EqSum, EqValue, Scalars_Out, Scalars_In) :-
	retrieve_coeff(Sum, EqVar, Coeff), 
	retrieve_coeff(EqSum, EqVar, EqCoeff), 
	compare(Op, 0, Coeff),
	compare(EqOp, 0, EqCoeff),
	multiplier_substitute_eq(Op, EqOp, Coeff, EqCoeff, Scale, EqScale),
	NewValue is Value * Scale + EqValue * EqScale,
	multiply_sum(Sum, Scale, ScaledSum),
	multiply_sum(EqSum, EqScale, ScaledEqSum),
	keysort(ScaledSum, SortedSum),
	keysort(ScaledEqSum, SortedEqSum),
	merge_sum(SortedSum, SortedEqSum, NewSum),
	tighten(RelOp, NewSum, NewValue, Scalars_Out, Scalars_In).

retrieve_coeff([X-Coeff | _], EqX, EqCoeff) :- 
	X == EqX, 
	!,
	EqCoeff = Coeff.
retrieve_coeff([_ | Sum], EqX, EqCoeff) :- 
	retrieve_coeff(Sum, EqX, EqCoeff).

merge_sum([], Sum, NewSum) :-
	!,
	NewSum = Sum.
merge_sum(Sum, [], NewSum) :-
	!,
	NewSum = Sum.
merge_sum([X-Coeff|Sum], [EqX-EqCoeff|EqSum], NewSum) :-
	compare(Op, X, EqX),
	merge_sum(Op, X, Coeff, Sum, EqX, EqCoeff, EqSum, NewSum).

merge_sum(<, X, Coeff, Sum, EqX, EqCoeff, EqSum, [X-Coeff | NewSum]) :-
	merge_sum(Sum, [EqX-EqCoeff | EqSum], NewSum).
merge_sum(=, X, Coeff, Sum, _, EqCoeff, EqSum, NewSum) :-
	NewCoeff is Coeff + EqCoeff,
	NewCoeff \== 0,
       	!,	
	NewSum = [X-NewCoeff|RestNewSum],
	merge_sum(Sum, EqSum, RestNewSum).
merge_sum(=, _, _, Sum, _, _, EqSum, NewSum) :-
	merge_sum(Sum, EqSum, NewSum).
merge_sum(>, X, Coeff, Sum, EqX, EqCoeff, EqSum, [EqX-EqCoeff | NewSum]) :-
	merge_sum([X-Coeff|Sum], EqSum, NewSum).

negative_sum(Sum, NegativeSum) :-
	multiply_sum(Sum, -1, NegativeSum).

multiply_sum([], _, []).
multiply_sum([X-Coeff | Sum], Multiplier, [X-ScaledCoeff | ScaledSum]) :-
	ScaledCoeff is Coeff * Multiplier,
	multiply_sum(Sum, Multiplier, ScaledSum).

divide_sum([], _, []).
divide_sum([X-Coeff | Sum], Divisor, [X-ScaledCoeff | ScaledSum]) :-
	ScaledCoeff is Coeff // Divisor,
	divide_sum(Sum, Divisor, ScaledSum).

gcd_sum([_-Coeff|Sum], GCD) :-
	PositiveCoeff is abs(Coeff),
	gcd_sum(Sum, PositiveCoeff, GCD).

gcd_sum([], GCD, GCD).
gcd_sum([_-Coeff|Sum], PartGCD, GCD) :-
	PositiveCoeff is abs(Coeff),
	gcd(PartGCD, PositiveCoeff, MiddleGCD),
	gcd_sum(Sum, MiddleGCD, GCD).

gcd(X, Y, GCD) :-
	X < Y,
	!,
	gcd(Y, X, GCD).
gcd(X, 0, GCD) :-
	!,
	GCD = X.
gcd(X, Y, GCD) :-
	Remain is X mod Y,
	gcd(Y, Remain, GCD).

multiplier_substitute_eq(<, <, Coeff, EqCoeff, Scale, EqScale) :- % checked on example
	!,
	Scale = EqCoeff,
	EqScale is -Coeff.
multiplier_substitute_eq(<, >, Coeff, EqCoeff, Scale, EqScale) :- % checked on example
	!,
	Scale is -EqCoeff,
	EqScale = Coeff.
multiplier_substitute_eq(>, <, Coeff, EqCoeff, Scale, EqScale) :- % checked on example
	!,
	Scale = EqCoeff,
	EqScale is -Coeff.
multiplier_substitute_eq(>, >, Coeff, EqCoeff, Scale, EqScale) :- % checked on example
	Scale is -EqCoeff,
	EqScale = Coeff.

substitute_ineqs_ineqs([], _, Project, Project).
substitute_ineqs_ineqs([Plus-Minus|Pairs], Var, Project, Residue) :-
	substitute_ineq_ineq(Minus, Plus, Var, PartProject, Residue),
	substitute_ineqs_ineqs(Pairs, Var, Project, PartProject).

substitute_ineq_ineq(scalar(Sum1,=<,Value1), scalar(Sum2,=<,Value2), Var, Scalars_Out, Scalars_In) :-
	retrieve_coeff(Sum1, Var, Coeff1), 
	retrieve_coeff(Sum2, Var, Coeff2), 
	Scale1 is abs(Coeff2),
	Scale2 is abs(Coeff1),
	NewValue is Value1 * Scale1 + Value2 * Scale2,
	multiply_sum(Sum1, Scale1, ScaledSum1),
	multiply_sum(Sum2, Scale2, ScaledSum2),
	keysort(ScaledSum1, SortedSum1),
	keysort(ScaledSum2, SortedSum2),
	merge_sum(SortedSum1, SortedSum2, NewSum),
	tighten_ineq(NewSum, NewValue, Scalars_Out, Scalars_In).

%inequality_size_bound(8).

tighten(=, NewSum, NewValue, Scalars_Out, Scalars_In) :-
	tighten_eq(NewSum, NewValue, Scalars_Out, Scalars_In).
tighten(=<, NewSum, NewValue, Scalars_Out, Scalars_In) :-
	tighten_ineq(NewSum, NewValue, Scalars_Out, Scalars_In).

tighten_eq([], Value, Scalars_Out, Scalars_In) :-
	Value == 0,
	!,
	Scalars_Out = Scalars_In.
tighten_eq([], Value, Scalars_Out, Scalars_In) :-  
	Value \== 0,
	!,
	Scalars_Out = [false | Scalars_In].
tighten_eq(Sum, Value, Scalars_Out, Scalars_In) :-
	!,
	Scalars_Out = [scalar(TightSum, =, TightValue) | Scalars_In],
	PositiveValue is abs(Value),
	(PositiveValue == 0 ->
		gcd_sum(Sum, GCD);
		gcd_sum(Sum, PositiveValue, GCD)),
	TightValue is Value // GCD,
	divide_sum(Sum, GCD, TightSum).
tighten_eq(Sum, Value, Scalars_Out, Scalars_In) :-
	Scalars_Out = [Scalar | Scalars_In],
	gcd_sum(Sum, GCD),
	compare(Op, Value, 0),
	tighten_eq_aux(Op, Sum, Value, GCD, Scalar).

tighten_eq_aux(<, Sum, Value, GCD, Scalar) :-
	0 is abs(Value) mod GCD,
	!,
	Scalar = scalar(TightSum, =, TightValue),
	NegativeGCD is -GCD,
	TightValue is Value // NegativeGCD,
	divide_sum(Sum, NegativeGCD, TightSum).
tighten_eq_aux(<, _, _, _, false).
tighten_eq_aux(=, Sum, Value, GCD, scalar(TightSum, =, TightValue)) :-
	TightValue is Value // GCD,
	divide_sum(Sum, GCD, TightSum).
tighten_eq_aux(>, Sum, Value, GCD, Scalar) :-
	0 is Value mod GCD,
	!,
	Scalar = scalar(TightSum, =, TightValue),
	TightValue is Value // GCD,
	divide_sum(Sum, GCD, TightSum).
tighten_eq_aux(>, _, _, _, false).

tighten_ineq([], Value, Scalars_Out, Scalars_In) :-
	Value < 0,
	!,
	Scalars_Out = [false | Scalars_In].
tighten_ineq([], Value, Scalars_Out, Scalars_In) :-
	Value >= 0,
	!,
	Scalars_Out = Scalars_In.
tighten_ineq(Sum, Value, Scalars_Out, Scalars_In) :-
	Scalars_Out = [scalar(TightSum, =<, TightValue) | Scalars_In],
	gcd_sum(Sum, GCD),
	divide_sum(Sum, GCD, TightSum),
	tighten_ineq_aux(Value, GCD, TightValue).

tighten_ineq_aux(Value, GCD, TightValue) :-
	Value >= 0,
	!,
       	TightValue is Value // GCD.
tighten_ineq_aux(Value, GCD, TightValue) :-
	0 is abs(Value) mod GCD,
	!,
       	TightValue is Value // GCD.
tighten_ineq_aux(Value, GCD, TightValue) :-
	TightValue is (Value // GCD) - 1.

my_subtract([], Vars, Vars).
my_subtract([TargetVar | TargetVars], Vars, Diff) :-
	delete(Vars, TargetVar, Residue),
	my_subtract(TargetVars, Residue, Diff).

sum_member(Sum, Var) :- \+ not_sum_member(Sum, Var).

not_sum_member([], _).
not_sum_member([Var-_ | Sum], Element) :-
	Var \== Element,
	not_sum_member(Sum, Element).

scale([], _, Acc, Acc). 
scale([scalar(Sum, RelOp, Value) | In], Sigma, Acc, Out) :- 
	Value == 0,
	!,
	scale(In, Sigma, [scalar(Sum, RelOp, 0) | Acc], Out).
scale([scalar(Sum, RelOp, Value) | In], Sigma, Acc, Out) :- 
	NegValue is -Value,
	scale(In, Sigma, [scalar([Sigma-NegValue | Sum], RelOp, 0) | Acc], Out).

add_vect([], [], [], Acc, Acc).
add_vect([U|Us], [V|Vs], [W|Ws], Acc, Out) :-
    	add_vect(Us, Vs, Ws, [scalar([W-(-1),U-1,V-1], =, 0) | Acc], Out).

%%% PRETTY PRINTING

scalars_to_goal([], true).
scalars_to_goal([Scalar | Scalars], Atoms) :-
	scalars_to_goal_aux(Scalars, Scalar, Atoms).

scalars_to_goal_aux([], Scalar, Atom) :-
	scalar_to_goal(Scalar, Atom).
scalars_to_goal_aux([Scalar | Scalars], Scalar1, (Atom, Atoms)) :-
	scalar_to_goal(Scalar1, Atom),
	scalars_to_goal([Scalar | Scalars], Atoms).

scalar_to_goal(ground(Var, Vars), Atom) :-
	!,
	Atom = ground_iff(Var, Vars).
scalar_to_goal(scalar(Sum, RelOp, Value), Atom) :-
	Value >= 0,
	!,
	sum_to_expr(Sum, 0, Value, LExpr, RExpr),
	Atom =.. [RelOp, LExpr, RExpr].
scalar_to_goal(scalar(Sum, RelOp, Value), Atom) :-
	!,
	NegValue is -Value,
	sum_to_expr(Sum, NegValue, 0, LExpr, RExpr),
	Atom =.. [RelOp, LExpr, RExpr].
scalar_to_goal(Atom, Atom).

sum_to_expr([], LAcc, RAcc, LAcc, RAcc).
sum_to_expr([X-Coeff | Sum], LAcc, RAcc, LExpr, RExpr) :-
	Coeff >= 0,
	!,
	form_sub_expr(Coeff, X, SubExpr),
	augment_sub_expr(LAcc, SubExpr, NewLAcc),
	sum_to_expr(Sum, NewLAcc, RAcc, LExpr, RExpr).
sum_to_expr([X-Coeff | Sum], LAcc, RAcc, LExpr, RExpr) :-
	Coeff < 0,
	NegCoeff is -Coeff,
	form_sub_expr(NegCoeff, X, SubExpr),
	augment_sub_expr(RAcc, SubExpr, NewRAcc),
	sum_to_expr(Sum, LAcc, NewRAcc, LExpr, RExpr).

form_sub_expr(Coeff, X, SubExpr) :-
	Coeff == 1,
	!,
	SubExpr = X.
form_sub_expr(Coeff, X, SubExpr) :-
	SubExpr = Coeff * X.

augment_sub_expr(Expr, SubExpr, NewExpr) :-
	Expr == 0,
	!,
	NewExpr = SubExpr.
augment_sub_expr(Expr, SubExpr, NewExpr) :-
	NewExpr = Expr + SubExpr.

projections(Scalars, SortedProject) :-
	term_variables(Scalars, Space),
	projections(Space, [], Scalars, [], [], Project),
	sort(Project, SortedProject).

residual([_Var | Vars], Vars).
residual([Var | Vars], [Var | ResidualVars]) :- 
	residual(Vars, ResidualVars).

projections([], _, []).
projections([Space-Scalars | ResidualPairs], PrevSpaces, Project) :-
	my_not_memberchk(PrevSpaces, Space),
	!,
	project(Scalars, Space, NewScalars),
	append(NewScalars, ResidualProject, Project),
	projections(Space, [], NewScalars, Pairs, [Space | PrevSpaces], ResidualProject).
projections([_ | ResidualPairs], PrevSpaces, Project) :-
	projections(ResidualPairs, PrevSpaces, Project).

projections([], _, _, Pairs, PrevSpaces, Project) :-
	projections(Pairs, PrevSpaces, Project).
projections([Var | Space], SpaceAcc, NewScalars, Pairs, PrevSpaces, Project) :-
	append(Space, SpaceAcc, NewSpace),
	projections(Space, [Var | SpaceAcc], NewScalars, [NewSpace-NewScalars|Pairs], PrevSpaces, Project).

%%% TEST DATA

test(0, project([L,I,F,C], Project)) :-
	project([scalar([A- -1],=<,0),
	scalar([B- -1],=<,0),
	scalar([A-1,B-1],=,1),
	scalar([C- -1,D-1,E-1],=,0),
	scalar([F- -1,G-1,H-1],=,0),
	scalar([I- -1,J-1,K-1],=,0),
	scalar([L- -1,M-1,N-1],=,0),
	scalar([B- -1,E-1],=<,0),
	scalar([E- -1],=<,0),
	scalar([H- -1,E-2],=<,0),
	scalar([N- -2,H-1,E-2],=<,0),
	scalar([B- -2,N- -1,H-1],=<,0),
	scalar([B- -2,N-3,K- -3,H- -1,E- -4],=<,0),
	scalar([B- -2,N-1,K- -1,E- -2],=<,0),
	scalar([A- -2,D-1],=<,0),
	scalar([A-1,D- -1],=<,0),
	scalar([G- -1,D-2],=<,0),
	scalar([M- -2,G-1,D-2],=<,0),
	scalar([A- -2,M- -1,G-1],=<,0),
	scalar([A- -2,M-3,J- -3,G- -1,D- -4],=<,0),
	scalar([A- -2,M-1,J- -1,D- -2],=<,0)], [L,I,F,C], Project).

test(1, project(Scalars, TargetVars, Project)) :-
	TargetVars = [X, Y],
	Scalars = [scalar([X-1, Z-(-3)], =, 2),
		scalar([Z-2, W-(-5)], =, 0),
		scalar([X-1, Y-1, W-1], =, 1),
		scalar([Z-1, Y-(-1)], =, 1)],
	project(Scalars, TargetVars, Project).
% answer Project = [scalar([X-17,Y-15],=,19), scalar([X-1,Y-(-3)],=,5)]

test(2, project(Scalars, TargetVars, Project)) :-
	TargetVars = [X],
	Scalars = [scalar([X-1, Y-(3)], =, 2),
		scalar([X-2, Y-(4)], =, 1)],
	project(Scalars, TargetVars, Project).
% answer Project = [false]

test(3, project(Scalars, TargetVars, Project)) :-
	TargetVars = [X],
	Scalars = [scalar([X-2, Y-(-2)], =, 2),
	scalar([X-4, Y-(-3)], =, 1)],
	project(Scalars, TargetVars, Project).
% answer Project = [scalar([X-(-1)],=,2)]

test(4, project(Scalars, TargetVars, Project)) :-
	TargetVars = [X],
	Scalars = [scalar([X-1, Y-(3)], =, 2),
		scalar([X-2, Y-(-4)], =, 1)],
	project(Scalars, TargetVars, Project).
% answer Project = [false]

test(5, project(Scalars, TargetVars, Project)) :-
	TargetVars = [X],
	Scalars = [scalar([X-1, Y-(3)], =<, 2),
		scalar([Y-6, Z-2], =<, 0),
		scalar([X-2, Z-(-3)], =<, 0),
		scalar([X-2, Y-(-4)], =<, 1)],
	project(Scalars, TargetVars, Project).
% answer Project = [scalar([X-1], =<, 0] 

test(6, convex_hull(Xs, ScalarsXs, Ys, ScalarsYs, Zs, ScalarsZs)) :-
	ScalarsXs = [scalar([X-1], =, 0), scalar([Y-1], =, 1)],
	ScalarsYs = [scalar([X-(-1)], =<, 0), scalar([X-1, Y-(-1)], =, 0)],
       	Xs = [X, Y],
       	Ys = [X, Y],
	convex_hull(Xs, ScalarsXs, Ys, ScalarsYs, Zs, ScalarsZs).
% answer Project = [scalar([X-1, Y-(-1)],=<,0), scalar([X-(-1), Y-1],=<,1), scalar([X-(-1)],=<,0)]

test(extra, convex_hull(Args, Scalars, Args, FixScalars, LubArgs, LubScalars)) :-
	Args = [A, B, C, D],
	Scalars = [scalar([A-1,B- -1,D- -2],=<,2),
		scalar([A-3,B- -3,C- -1,D- -4],=<,2),
		scalar([A- -1,C-1],=<,2),
		scalar([A- -2,C-1,D-2],=<,0),
		scalar([C- -1,D-2],=<,0),
		scalar([D- -1],=<,0),
		scalar([D-1],=<,1)],
	FixScalars = [scalar([A-1,B- -1,D- -2],=<,2),
		scalar([A-3,B- -3,C- -1,D- -4],=<,2),
		scalar([A- -1,C-1],=<,2),
		scalar([A- -2,C-1,D-2],=<,0),
		scalar([C- -1,D-2],=<,0),
		scalar([D- -1],=<,-1),
		scalar([D-1],=<,2)],
	convex_hull(Args, Scalars, Args, FixScalars, LubArgs, LubScalars).

test(7, convex_hull(Xs, ScalarsXs, Ys, ScalarsYs, Zs, ScalarsZs)) :-
	ScalarsXs = [scalar([X-1], =, 0), scalar([Y-1], =, 1)],
	ScalarsYs = [scalar([X-1], =, 2), scalar([Y-1], =, 3)],
       	Xs = [X, Y],
       	Ys = [X, Y],
	convex_hull(Xs, ScalarsXs, Ys, ScalarsYs, Zs, ScalarsZs).
% answer Project = [scalar([Y-1],=<,3), scalar([Y-(-1)],=<,-1), scalar([X-(-1), Y-1],=,1)]

test(8, unsatisfiable(Scalars)) :-
	Scalars = [scalar([X-1, Y-(-1)], =, 0), scalar([X-1, Y-2], =, 2)],
	unsatisfiable(Scalars).
% true

test(9, satisfiable(Scalars)) :-
	Scalars = [scalar([X-1, Y-(-1)], =, 0), scalar([X-1, Y-2], =<, 7)],
	\+ unsatisfiable(Scalars).
% false

test(10, entails(Scalars1, Scalars2)) :-
	Scalars1 = [scalar([X-1], =, 1), scalar([Y-1], =, 2)],
	Scalars2 = [scalar([X-(-1)], =<, 0), scalar([X-2, Y-(-1)], =, 0)],
	entails(Scalars1, Scalars2).
% true

test(11, entails(Scalars1, Scalars2)) :-
	Scalars1 = [scalar([X-1], =, 1), scalar([Y-1], =, 2)],
	Scalars2 = [scalar([Y-1], =, 2), scalar([X-1], =, 1)],
	entails(Scalars1, Scalars2).
% true

test(12, entails(Scalars1, Scalars2)) :-
	Scalars1 = [scalar([X-(-1)], =<, 0), scalar([X-2, Y-(-1)], =, 0)],
	Scalars2 = [scalar([X-1], =, 1), scalar([Y-1], =, 2)],
	entails(Scalars1, Scalars2).
% false

test(13, convex_hull(Xs, ScalarsXs, Ys, ScalarsYs, Zs, ScalarsZs)) :-
       	Xs = [A, B, C, D],
       	Ys = Xs,
	ScalarsXs = [scalar([A-1, D-(-1)], =, 1),
		scalar([C-1], =, 0),
		scalar([B-(-1)], =<, 0),
		scalar([D-(-1)], =<, 0),
		scalar([D-1], =<, 1)],
	ScalarsYs = [scalar([A-1, D-(-1)], =, 0),
		scalar([C-1], =, 0),
		scalar([B-(-1)], =<, 0),
		scalar([D-(-1)], =<, 0),
		scalar([D-1], =<, 1)],
	convex_hull(Xs, ScalarsXs, Ys, ScalarsYs, Zs, ScalarsZs),
	Zs = Xs.

test(14, compress(Scalars, DenseScalars)) :-
	Scalars = [scalar([B-(-1)], =<, 0),
		scalar([B-(-1)], =<, 1),	
		scalar([C-1, B-(-1)], =<, 0),
		scalar([B-1, C-(-1)], =<, 3),
		scalar([C-(-1)], =<, 2),
		scalar([A-1, C-1, B-(-1)], =, 0),
		scalar([A-(-1)], =<, 0),
		scalar([A-1], =<, 1),
		scalar([B-(-1)], =<, 0),
		scalar([B-1, A-(-1), C-(-1)], =, 0)],
	compress(Scalars, DenseScalars).

test(15, project(Scalars, Args, Project)) :-
	Scalars = [scalar([D-(-1),E-(-1),A-1],=,2),
		scalar([B-1],=,0),
		scalar([E-1],=,0),
		scalar([D-(-1),C-1],=<,-1),
		scalar([],=,0),
		scalar([D-(-1)],=<,-1),
		scalar([D-(-1)],=<,0),
		scalar([E-(-1)],=<,0),
		scalar([B-(-1)],=<,0),
		scalar([C-(-1)],=<,0),
		scalar([E-(-1)],=<,0),
		scalar([A-(-1)],=<,0)],
	Args    = [A,B,C],
	project(Scalars, Args, Project).

test(16, compress(Scalars, DenseScalars)) :-
       Scalars = [scalar([A-(-1)],=<,0),
       		scalar([A-1,B-(-1)],=<,-3),
		scalar([C-(-1)],=<,0),
		scalar([B-(-1)],=<,-3),
		scalar([B-(-1)],=<,-2),
		scalar([B-(-1)],=<,0),
		scalar([],=,0),
		scalar([C-1],=,0)],
	compress(Scalars, DenseScalars).

test(17, widen(Scalars1, Scalars2, WideScalars)) :-
	Scalars1 = [scalar([X-1], =, 1), scalar([Y-1], =, 2)],
	Scalars2 = [scalar([Y-1], =, 2), scalar([X-1], =, 0)],
	widen(Scalars1, Scalars2, WideScalars).
% WideScalars = [scalar([Y-1], =, 2)]

test(18, project(Scalars, Args, Project, Time)) :-
	statistics(runtime, _),
	Args = [C,D,E,F,G,H,I,J,K,L,M,N,O,P],
	Scalars = [scalar([F-1],=,0),
		scalar([M-1,B- -1],=<,4),
		scalar([O- -1,A-1],=<,4),
		scalar([M- -2,B-2,O- -1,A-1],=<,0),
		scalar([M- -1,B-1],=<,0),
		scalar([A- -1],=<,-4),
		scalar([A- -2,B- -4,P-2,N-4,J-2,I- -9],=<,-14),
		scalar([A- -1,B- -2,P-1,N-2,J-1],=<,2),
		scalar([A- -2,B- -4,P-2,N-4,J- -7],=<,-14),
		scalar([A- -1,B- -2,P-1,N-2,J-1,H- -9],=<,-7),
		scalar([A- -2,P-2,J-2,I- -13],=<,-14),
		scalar([A- -1,P-1,J-1],=<,6),
		scalar([A- -2,P-2,J- -11],=<,-14),
		scalar([B-2,N- -2,I- -3,H- -2],=<,0),
		scalar([A- -2,B-4,P-2,N- -4,J-2,I- -15,H- -4],=<,-14),
		scalar([B-2,N- -2,J- -3,H- -2],=<,0),
		scalar([A- -2,B-4,P-2,N- -4,J- -13,H- -4],=<,-14),
		scalar([B-1,N- -1,H- -1],=<,3),
		scalar([A- -1,B-2,P-1,N- -2,J-1,H- -2],=<,8),
		scalar([B-1,N- -1,H- -4],=<,0),
		scalar([A- -1,B-2,P-1,N- -2,J-1,H- -17],=<,-7),
		scalar([A- -1,B- -2,N-2,J-1,H- -5],=<,-7),
		scalar([A- -11,B- -12,N-12,J-11,H- -75],=<,-77),
		scalar([A- -20,B-18,P-9,N- -18,J-20,H- -252],=<,-140),
		scalar([A- -2,P- -3,J-2,H- -18],=<,-14),
		scalar([A- -2,J- -7],=<,-14),
		scalar([A- -1,J-1],=<,2),
		scalar([A- -2,J-2,I- -9],=<,-14),
		scalar([A- -2,B- -4,N-4,J-2,I- -5],=<,-14),
		scalar([A- -1,B- -2,N-2,J-1],=<,-2),
		scalar([A- -2,B- -4,N-4,J- -3],=<,-14),
		scalar([B- -1,N-1],=<,0),
		scalar([L-1,K- -1,G- -1,D- -1,C- -1],=,7),
		scalar([C- -1],=<,0),
		scalar([D- -1],=<,0),
		scalar([E- -1],=<,0),
		scalar([G- -1],=<,0),
		scalar([I- -1],=<,0),
		scalar([J- -1],=<,0),
		scalar([K- -1],=<,0),
		scalar([N- -1],=<,0),
		scalar([P- -1],=<,0)],
	project(Scalars, Args, Project),
	statistics(runtime, [_, Time]).


