:- module(linear_fixpoint, [linear_lfp_program/2, linear_lfp_pretty_print/1]).
:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(polyhedra).
:- use_module(utilities).

:- dynamic linear_atom/3.
:- dynamic redo/0.

linear_lfp_program(SCCs, Status) :-
	retractall(redo),
	retractall(linear_atom(_, _, _)),
	lfp_sccs(SCCs, Status).

lfp_sccs([], _).
lfp_sccs([SCC | SCCs], Status) :-
	format("~nSCC = ~w ", [SCC]), ttyflush,
	lfp_scc(SCC, Status),
	lfp_sccs(SCCs, Status).

lfp_scc(SCC, status(Trace, K)) :-
	lfp_scc(SCC, Trace, K).

lfp_scc(Index, Trace, _K) :-
	integer(Index),
	!,
	lfp_scc_pass([Index], Trace).
%lfp_scc(Indices, Trace, K) :-
%	K > 1,
%	!,
%	lfp_scc_pass(Indices, Trace),
%	K1 is K - 1,
%	lfp_scc(Indices, Trace, K1).
lfp_scc(Indices, Trace, K) :-
	indices_to_names(Indices, [], Pairs),
	read_state(Pairs, Atoms, LastState),
	lfp_scc_pass(Indices, Trace),
	read_state(Pairs, Atoms, NextState),
	\+ entails_state(NextState, LastState),
	!,
%	widen_state(LastState, NextState, Atoms, Trace, WidenState),
%	write_state(Atoms, WidenState),
	lfp_scc(Indices, Trace, K).
lfp_scc(_, _, _).

display_state([], []).
display_state([Atom | Atoms], [Scalars | State]) :-
	scalars_to_goal(Scalars, Body),
	portray_clause((Atom :- Body)),
	display_state(Atoms, State).

indices_to_names([], Pairs, Pairs).
indices_to_names([Index | Indices], Acc, Pairs) :-
        abstract:head_to_linear_clause(Index, Atom, _),
       	functor(Atom, Name, Arity),
	Pair = Name/Arity,
	(memberchk(Pair, Acc) ->
		indices_to_names(Indices, Acc, Pairs);
		indices_to_names(Indices, [Pair | Acc], Pairs)).
		
read_state([], [], []).
read_state([Name/Arity | Pairs], [Atom | Atoms], [State | States]) :-
	functor(Atom, Name, Arity),
	linear_atom(Atom, State, []),
	read_state(Pairs, Atoms, States).

write_state([], []).
write_state([Atom | Atoms], [State | States]) :-
	retract(linear_atom(Atom, _, _)),
	append(State, Scalars_In, Scalars_Out),
	assert(linear_atom(Atom, Scalars_Out, Scalars_In)),
	write_state(Atoms, States).

entails_state([], []).
entails_state([Scalars | States], [FixScalars | FixStates]) :-
	entails(Scalars, FixScalars),
	entails_state(States, FixStates).

widen_state([], [], [], _, []).
widen_state([Scalars | States], [FixScalars | FixStates], [Atom | Atoms],
		Trace, [CompressedScalars | WidenStates]) :-
	widen(Scalars, FixScalars, WideScalars),
	compress(WideScalars, CompressedScalars),
	stepper_aux(Trace, fix_widen, Atom, Scalars),
	stepper_aux(Trace, new_widen, Atom, FixScalars),
	stepper(Trace, result_widen, Atom, CompressedScalars),
	widen_state(States, FixStates, Atoms, Trace, WidenStates).

lfp_scc_pass([], _).
lfp_scc_pass([Index | Indices], Trace) :-
	!,
	format("~nIndex = ~w", [Index]), ttyflush,
	(lfp_clause(Index, Trace); true),
	lfp_scc_pass(Indices, Trace).

lfp_clause(Index, Trace) :-
        abstract:head_to_linear_clause(Index, Atom, Atoms), 
        lfp_atoms(Atoms, Scalars),
	Atom =.. [_ | Args],
	project(Scalars, Args, Project),
        uncond_assert(Atom, Project, Trace),
        fail.

lfp_atoms([], []).
lfp_atoms([Atom | Atoms], [Atom | Scalars]) :- 
	functor(Atom, scalar, 3),
	!,
	lfp_atoms(Atoms, Scalars).
lfp_atoms([Atom | Atoms], Scalars) :- 
	linear_atom(Atom, Scalars, RestScalars),
	lfp_atoms(Atoms, RestScalars).

uncond_assert(Atom, Scalars, Trace) :-
	linear_atom(Atom, FixScalars, []), 
	!,
	Atom =.. [Name | Args],
	stepper_aux(Trace, fix_lub, Atom, FixScalars),
	stepper_aux(Trace, new_lub, Atom, Scalars),
	convex_hull(Args, Scalars, Args, FixScalars, LubArgs, LubScalars),
	LubAtom =.. [Name | LubArgs],
	LubAtom = Atom,
%	stepper_aux(Trace, result_lub, Atom, LubScalars),
	widen_scalars(LubScalars, 4, WidenScalars),
	stepper(Trace, widen_result_lub, Atom, WidenScalars),
	retract(linear_atom(Atom, _, _)),
	append(WidenScalars, Scalars_In, Scalars_Out),
	asserta(linear_atom(Atom, Scalars_Out, Scalars_In)). 
uncond_assert(Atom, Scalars, Trace) :-
	stepper(Trace, new, Atom, Scalars),
	append(Scalars, Scalars_In, Scalars_Out),
	asserta(linear_atom(Atom, Scalars_Out, Scalars_In)). 

widen_scalars([], _, []).
widen_scalars([scalar(Sum, RelOp, Value) | Scalars], Threshold, WidenScalars) :-
	abs(Value) =< Threshold,
	not_widen_sum(Sum, Threshold),
	!,
	WidenScalars = [scalar(Sum, RelOp, Value) | RestWidenScalars],
	widen_scalars(Scalars, Threshold, RestWidenScalars).
widen_scalars([_ | Scalars], Threshold, WidenScalars) :-
	widen_scalars(Scalars, Threshold, WidenScalars).

not_widen_sum([], _).
not_widen_sum([_-Coeff | Sum], Threshold) :-
	abs(Coeff) =< Threshold,
	not_widen_sum(Sum, Threshold).

stepper(off, _, _, _).
stepper(step, Message, Atom, Scalars) :-
	stepper_aux(log, Message, Atom, Scalars),
	read(Action),
	stepper_aux_aux(Action).
stepper(log, Message, Atom, Scalars) :-
	stepper_aux(log, Message, Atom, Scalars).

stepper_aux(off, _, _, _).
stepper_aux(step, Message, Atom, Scalars) :-
	stepper_aux(log, Message, Atom, Scalars).
stepper_aux(log, Message, Atom, Scalars) :-
	format("~n~w: ", [Message]), 
	scalars_to_goal(Scalars, Body),
	portray_clause((Atom :- Body)),
	ttyflush.

stepper_aux_aux(trace) :-
	!,
	trace.
stepper_aux_aux(abort) :-
	!,
	abort.
stepper_aux_aux(_).

%% PRETTY PRINTING CODE

linear_lfp_pretty_print(File) :-
	open(File, write, Stream),
	linear_lfp_pretty_print_aux(Stream).

linear_lfp_pretty_print_aux(Stream) :-
	linear_atom(Head, Scalars, []),
	scalars_to_goal(Scalars, Body),
	portray_clause(Stream, (Head :- Body)),
	fail.
linear_lfp_pretty_print_aux(Stream) :-
	close(Stream).




